# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment

from helpers.find_free_port import find_free_port
from helpers.launch_cogment import launch_model_registry, launch_directory

import pytest
import asyncio
import logging

logger = logging.getLogger("cogment.unit-tests")


@pytest.fixture(scope="function")
def model_registry_test_setup(test_cogment_app_dir, cogment_path):
    # Launch the directory
    directory_port = find_free_port()
    terminate_directory = launch_directory(
        app_directory=test_cogment_app_dir,
        directory_port=directory_port,
        cogment_path=cogment_path,
    )
    directory_endpoint = f"grpc://localhost:{directory_port}"

    model_registry_port = find_free_port()
    # Launch the model registry
    terminate_model_registry = launch_model_registry(
        app_directory=test_cogment_app_dir,
        model_registry_port=model_registry_port,
        cogment_path=cogment_path,
        directory_endpoint=directory_endpoint,
    )
    # Execute the test
    yield {
        "directory_endpoint": cogment.Endpoint(directory_endpoint),
    }

    # Terminate cogment
    terminate_model_registry()
    terminate_directory()


@pytest.mark.use_cogment
@pytest.mark.asyncio
async def test_model_registry(
    model_registry_test_setup, unittest_case, cog_settings, data_pb2
):
    context = cogment.Context(
        cog_settings=cog_settings,
        user_id="test_model_registry",
        prometheus_registry=None,
        directory_endpoint=model_registry_test_setup["directory_endpoint"]
    )

    await asyncio.sleep(5)
    model_registry = await context.get_model_registry()

    model_id = "test"
    model_user_data = {"model_key": "model_value"}
    version_user_data = {"version_key": "version_value"}

    # Use proto message as test 'model' due to ease of serializing/deserializing
    deserialized_model = data_pb2.Observation(observed_value=1)
    serialized_model = deserialized_model.SerializeToString()

    model = cogment.Model(
        model_id=model_id,
        serialized_model=serialized_model,
        user_data=model_user_data,
        version_user_data=version_user_data,
    )

    await asyncio.sleep(1)
    await model_registry.store_initial_version(model)
    retrieved_model = await model_registry.retrieve_version(model_id)

    assert retrieved_model.serialized_model == serialized_model
    assert retrieved_model.user_data == model_user_data
    assert retrieved_model.version_user_data == version_user_data
    assert retrieved_model.stored_version_info.data_size == len(serialized_model)
    assert retrieved_model.stored_version_info.version_number == 1
    assert retrieved_model.id == model_id
    assert retrieved_model.deserialized_model is None

    model.deserialized_model = deserialized_model
    await model_registry.store_version(model)
    retrieved_model = await model_registry.retrieve_version(model_id)

    assert retrieved_model.serialized_model is None
    assert retrieved_model.stored_version_info.version_number == 2
    assert retrieved_model.deserialized_model == deserialized_model

    def deserialize_function(data):
        obs = data_pb2.Observation()
        obs.ParseFromString(data)
        return obs

    model_registry._data_cache.clear()
    model_registry._info_cache.clear()
    retrieved_model = await model_registry.retrieve_version(
        model_id=model_id,
        version_number=-1,
        deserialize_func=deserialize_function
    )

    assert retrieved_model.serialized_model == serialized_model
    assert retrieved_model.stored_version_info.version_number == 2
    assert retrieved_model.deserialized_model == deserialized_model
    assert len(model_registry._data_cache) == 1
    assert len(model_registry._info_cache) == 1



# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment


def test_arbitrary_metadata(unittest_case):
    metadata = cogment.GrpcMetadata()
    metadata.add("First", "Value")
    metadata.add("second", "second")
    metadata.add("foO-bIn", b"blop")

    unittest_case.assertCountEqual(
        metadata.to_grpc_metadata(),
        [("first", "Value"), ("second", "second"), ("foo-bin", b"blop")],
    )


def test_metadata_copy(unittest_case):
    metadata = cogment.GrpcMetadata()
    metadata.add("First", "Value")

    metadata_copy = metadata.copy()

    unittest_case.assertCountEqual(
        metadata.to_grpc_metadata(),
        metadata_copy.to_grpc_metadata(),
    )

    metadata_copy.add("bLOp", "bloupy")

    unittest_case.assertCountEqual(
        metadata.to_grpc_metadata(),
        [("first", "Value")],
    )

    unittest_case.assertCountEqual(
        metadata_copy.to_grpc_metadata(),
        [("first", "Value"), ("blop", "bloupy")],
    )


def test_bad_binary_metadata(unittest_case):
    metadata = cogment.GrpcMetadata()
    with unittest_case.assertRaises(cogment.CogmentError):
        metadata.add("test", b"blop")


def test_bad_metadata_prefix(unittest_case):
    metadata = cogment.GrpcMetadata()
    with unittest_case.assertRaises(cogment.CogmentError):
        metadata.add("GrpC-test", "blop")


def test_basic_auth(unittest_case):
    metadata = cogment.GrpcMetadata()
    metadata.add("First", "Value")
    metadata.add_http_basic_auth("user", "password")

    unittest_case.assertCountEqual(
        metadata.to_grpc_metadata(),
        [
            ("first", "Value"),
            (
                "authorization",
                "basic dXNlcjpwYXNzd29yZA==",  # `echo -n "user:password" | base64``
            ),
        ],
    )



# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from copy import deepcopy
from typing import Tuple, List, Sequence, Union

from typing_extensions import Self
from cogment.errors import CogmentError

GrpcMetadataValue = Union[str, bytes]


class GrpcMetadata:
    """
    Metadata sent alongside gRPC calls to Cogment endpoints

    Some specifications from https://grpc.io/docs/guides/metadata/

    """

    def __init__(self) -> None:
        self._metadata: List[Tuple[str, GrpcMetadataValue]] = []

    def copy(self) -> Self:
        """Return a deepcopy of the metadata, useful to extend w/o side effects"""
        return deepcopy(self)

    def add(self, key: str, value: GrpcMetadataValue):
        """Add arbitrary metadata key/value pair"""
        key = key.lower()

        if key.startswith("grpc-"):
            raise CogmentError("'grpc-' metadata prefix is reserved by grpc itself")

        if key.endswith("-bin"):
            if not isinstance(value, bytes):
                raise CogmentError("Key ending with '-bin' should have a bytes value")
            # Unclear what exactly should be done for binary values
            # should they be encoded in base64? the official python example says no
            # cf. https://github.com/grpc/grpc/blob/master/examples/python/metadata/metadata_client.py
            self._metadata.append((key, value))

        else:
            if isinstance(value, bytes):
                raise CogmentError("Binary metadata should have a key ending with '-bin'")

            self._metadata.append((key, value))

    def add_http_basic_auth(self, username: str, password: str) -> None:
        """Add http basic auth 'authorization' metadata item"""
        bin_auth: bytes = f"{username}:{password}".encode("utf-8")
        base64_auth: str = base64.b64encode(bin_auth).decode("utf-8")
        self._metadata.append(("authorization", f"basic {base64_auth}"))

    def to_grpc_metadata(self) -> Sequence[Tuple[str, Union[str, bytes]]]:
        """Return the metadata ready to be sent alongside a gRPC request"""
        return self._metadata



# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import urllib.request
from unittest.mock import AsyncMock

import pytest

import cogment

from helpers.find_free_port import find_free_port
from helpers.launch_cogment import launch_orchestrator

logger = logging.getLogger("cogment.unit-tests")


@pytest.fixture(scope="function")
def cogment_test_setup(test_cogment_app_dir, cogment_path):
    orchestrator_port = find_free_port()
    test_port = find_free_port()
    # Launch the orchestrator
    terminate_orchestrator = launch_orchestrator(
        test_cogment_app_dir,
        cogment_path=cogment_path,
        orchestrator_port=orchestrator_port,
        test_port=test_port
    )

    # Execute the test
    yield {
        "orchestrator_endpoint": f"grpc://localhost:{orchestrator_port}",
        "test_port": test_port,
    }

    # Terminate orchestrator
    terminate_orchestrator()


# Base implementations
def create_pre_trial_hook(environment_implementation):
    async def _pre_trial_hook(session):
        session.environment_implementation = environment_implementation
        session.validate()
    return _pre_trial_hook


class TestIntegration:
    @pytest.mark.use_cogment
    @pytest.mark.asyncio
    async def test_environment_controlled_trial(
        self, cogment_test_setup, unittest_case, cog_settings, data_pb2
    ):
        controller_trial_id = None
        target_tick_count = 10
        environment_call_count = 0
        environment_tick_count = 0
        environment_trial_id = None

        async def environment(environment_session):
            nonlocal environment_trial_id
            nonlocal environment_call_count
            nonlocal environment_tick_count

            assert environment_session.config is not None
            assert environment_session.config.env_config_value == 0

            environment_trial_id = environment_session.get_trial_id()
            environment_call_count += 1

            environment_session.start([("*", data_pb2.Observation(observed_value=12))])

            async for event in environment_session.event_loop():
                if event.type == cogment.EventType.FINAL:
                    continue

                assert len(event.actions) == len(
                    environment_session.get_active_actors()
                )
                environment_tick_count += 1

                if environment_tick_count == 1:
                    logger.debug(f"--Sending message and rewards to Actors")

                    actors = environment_session.get_active_actors()
                    for actor in actors:
                        user_data = data_pb2.MyFeedbackUserData(
                            a_bool=False, a_float=3.0
                        )
                        environment_session.add_reward(
                            21.0, 1.0, [actor.actor_name], user_data=user_data
                        )

                        msg = data_pb2.MyMessageUserData(
                            a_string="A personal string", an_int=21
                        )
                        environment_session.send_message(msg, [actor.actor_name])

                    msg = data_pb2.MyMessageUserData(
                        a_string="An universal string", an_int=42
                    )
                    environment_session.send_message(msg, ["*"])

                if environment_tick_count >= target_tick_count:
                    environment_session.end(
                        [("*", data_pb2.Observation(observed_value=12))]
                    )
                else:
                    environment_session.produce_observations(
                        [("*", data_pb2.Observation(observed_value=12))]
                    )

        pre_hook_called_count = 0

        async def pre_trial_hook(session):
            nonlocal pre_hook_called_count

            assert session.trial_config is not None
            assert session.trial_config.trial_config_value == 31
            assert session.trial_config.env_config is not None

            session.environment_config = session.trial_config.env_config

            session.actors.append(
                {
                    "name": "client_1",
                    "actor_class": "my_actor_class_1",
                    "endpoint": "client",
                    "implementation": "",
                    "config": None,
                }
            )

            pre_hook_called_count += 1

            session.validate()

        agent_call_count = 0
        agent_observation_count = 0
        agent_final_data_count = 0
        agent_message_count = 0
        had_universal_message = False
        had_personal_message = False
        had_message_from_client = 0
        had_reward_from_client = 0
        agent_trial_id = {}
        total_reward = 0
        agents_ended = {}
        agents_ended["actor_1"] = asyncio.get_running_loop().create_future()
        agents_ended["actor_2"] = asyncio.get_running_loop().create_future()

        async def agent(actor_session):
            nonlocal agent_call_count
            nonlocal agent_observation_count
            nonlocal agent_final_data_count
            nonlocal agent_message_count
            nonlocal had_universal_message
            nonlocal had_personal_message
            nonlocal had_message_from_client
            nonlocal had_reward_from_client
            nonlocal total_reward
            nonlocal agents_ended
            nonlocal agent_trial_id
            assert actor_session.name in agents_ended
            agent_trial_id[actor_session.name] = actor_session.get_trial_id()
            agent_call_count += 1

            actor_session.start()

            last_tick_id = 0
            async for event in actor_session.event_loop():
                if event.observation:
                    assert event.observation.snapshot.observed_value == 12
                    assert last_tick_id == event.observation.tick_id
                    last_tick_id += 1

                    actor_session.do_action(data_pb2.Action(action_value=-1))
                    agent_observation_count += 1

                for message in event.messages:
                    msg = data_pb2.MyMessageUserData()
                    msg.ParseFromString(message.payload.value)
                    if message.sender_name == "client_1":
                        assert actor_session.name == "actor_2"
                        had_message_from_client += 1
                    elif msg.an_int == 42:
                        had_universal_message = True
                        agent_message_count += 1
                    elif msg.an_int == 21:
                        had_personal_message = True
                        agent_message_count += 1

                for reward in event.rewards:
                    for src in reward.all_sources():
                        if src.sender_name == "client_1":
                            had_reward_from_client += 1
                            assert actor_session.name == "actor_1"
                            assert src.value == 30
                            assert src.confidence == 0.5
                        else:
                            total_reward += src.value
                            assert src.value == 21.0

                if event.type == cogment.EventType.FINAL:
                    agent_final_data_count += 1

            agents_ended[actor_session.name].set_result(True)

        client_call_count = 0
        client_observation_count = 0
        client_final_data_count = 0
        client_message_count = 0
        client_had_universal_message = False
        client_had_personal_message = False
        client_total_reward = 0
        agents_ended["client_1"] = asyncio.get_running_loop().create_future()

        async def client_actor(actor_session):
            nonlocal client_call_count
            nonlocal client_observation_count
            nonlocal client_final_data_count
            nonlocal client_message_count
            nonlocal client_had_universal_message
            nonlocal client_had_personal_message
            nonlocal client_total_reward
            nonlocal agents_ended

            assert actor_session.name in agents_ended
            agent_trial_id[actor_session.name] = actor_session.get_trial_id()
            client_call_count += 1

            actor_session.start()

            last_tick_id = 0
            async for event in actor_session.event_loop():
                if event.observation:
                    assert event.observation.snapshot.observed_value == 12
                    assert last_tick_id == event.observation.tick_id
                    last_tick_id += 1

                    actor_session.do_action(data_pb2.Action(action_value=-1))
                    client_observation_count += 1

                    msg = data_pb2.MyMessageUserData(
                        a_string="A message from aclient", an_int=2
                    )
                    actor_session.send_message(msg, ["actor_2"])

                    actor_session.add_reward(30.0, 0.5, ["actor_1"])

                for message in event.messages:
                    msg = data_pb2.MyMessageUserData()
                    msg.ParseFromString(message.payload.value)
                    client_message_count += 1
                    if msg.an_int == 42:
                        client_had_universal_message = True
                    elif msg.an_int == 21:
                        client_had_personal_message = True

                for reward in event.rewards:
                    client_total_reward += reward.value
                    assert reward.value == 21.0

                if event.type == cogment.EventType.FINAL:
                    client_final_data_count += 1

            agents_ended[actor_session.name].set_result(True)

        context = cogment.Context(
            cog_settings=cog_settings, user_id="test_environment_controlled_trial"
        )

        context.register_environment(impl=environment)
        context.register_actor(impl_name="test", impl=agent)
        context.register_actor(
            impl_name="client",
            impl=client_actor,
            actor_classes=[
                "my_actor_class_1",
            ],
        )
        context.register_pre_trial_hook(pre_trial_hook)

        prometheus_port = find_free_port()
        served_endp = cogment.ServedEndpoint(cogment_test_setup["test_port"])
        asyncio.create_task(
            context.serve_all_registered(
                served_endpoint=served_endp, prometheus_port=prometheus_port
            )
        )
        await asyncio.sleep(1)

        endp = cogment.Endpoint(cogment_test_setup["orchestrator_endpoint"])
        controller = context.get_controller(endpoint=endp)

        trial_ended = asyncio.get_running_loop().create_future()

        async def state_tracking(trial_controller):
            nonlocal trial_ended
            logger.debug(f"--Iterating over trial state...")

            state_count = 0
            state_trial_id = []
            state_itor = trial_controller.watch_trials()
            async for info in state_itor:
                logger.debug(f"--Trial state: {info}")
                assert info.trial_id != 0
                assert info.state != cogment.TrialState.UNKNOWN

                if info.state == cogment.TrialState.INITIALIZING:
                    state_trial_id.append(info.trial_id)
                    state_count += 1
                if info.state == cogment.TrialState.PENDING:
                    state_trial_id.append(info.trial_id)
                    state_count += 10
                if info.state == cogment.TrialState.RUNNING:
                    state_trial_id.append(info.trial_id)
                    state_count += 100
                if info.state == cogment.TrialState.TERMINATING:
                    state_trial_id.append(info.trial_id)
                    state_count += 1000
                if info.state == cogment.TrialState.ENDED:
                    state_trial_id.append(info.trial_id)
                    state_count += 10000
                    break

            await state_itor.aclose()
            trial_ended.set_result((state_count, state_trial_id))
            logger.debug(f"--Finished iterating over trial state: trial [{controller_trial_id}] ended")

        asyncio.create_task(state_tracking(controller))
        await asyncio.sleep(1)

        controller_trial_id = await controller.start_trial(
            trial_config=data_pb2.TrialConfig(trial_config_value=31)
        )
        await context.join_trial(
            trial_id=controller_trial_id, endpoint=endp, impl_name="client"
        )

        await trial_ended
        logger.info("--State reported trial ended")

        count, ids = trial_ended.result()
        assert count == 11111
        assert len(ids) == 5
        for id in ids:
            assert id == controller_trial_id

        # Although the trial has ended, the agents could still be processing the last observation
        assert len(agents_ended) == 3
        await asyncio.wait(agents_ended.values())

        prometheus_connection = urllib.request.urlopen(
            "http://localhost:" + str(prometheus_port)
        )
        promethus_bytes = prometheus_connection.read()
        promethus_data = promethus_bytes.decode("utf8")
        prometheus_connection.close()

        index = promethus_data.find("actor_decide_processing_seconds_count")
        assert index != -1
        index = promethus_data.find('impl_name="test"')
        assert index != -1
        index = promethus_data.find('name="actor_1"')
        assert index != -1
        index = promethus_data.find('name="actor_2"')
        assert index != -1
        index = promethus_data.find("environment_trial_duration_in_second_count")
        assert index != -1
        index = promethus_data.find('impl_name="default"')
        assert index != -1
        index = promethus_data.find("Can you find anything ?")
        assert index == -1

        assert environment_trial_id is not None
        assert controller_trial_id is not None
        assert "actor_1" in agent_trial_id
        assert "actor_2" in agent_trial_id
        assert controller_trial_id == environment_trial_id
        assert controller_trial_id == agent_trial_id["actor_1"]
        assert controller_trial_id == agent_trial_id["actor_2"]

        assert environment_call_count == 1
        assert environment_tick_count == target_tick_count
        assert agent_call_count == 2
        assert agent_observation_count / agent_call_count == target_tick_count + 1
        assert agent_final_data_count == 2
        assert had_universal_message
        assert had_personal_message
        assert agent_message_count == 4
        assert total_reward == 42.0
        assert had_message_from_client == 10
        assert had_reward_from_client == 10

        assert client_call_count == 1
        assert client_observation_count == target_tick_count + 1
        assert client_final_data_count == 1
        assert client_had_universal_message
        assert client_had_personal_message
        assert client_message_count == 2
        assert client_total_reward == 21.0

        assert pre_hook_called_count == 1

        logger.info(f"test_environment_controlled_trial finished")
        await context._grpc_server.stop(grace=5.0)  # To prepare for next test

    @pytest.mark.use_cogment
    @pytest.mark.asyncio
    async def test_controller_controlled_trial(
        self, cogment_test_setup, unittest_case, cog_settings, data_pb2
    ):
        environment_call_count = 0
        environment_tick_count = 0

        async def environment(environment_session):
            nonlocal environment_call_count
            nonlocal environment_tick_count

            environment_call_count += 1
            environment_tick_count += 1
            environment_session.start(
                [("*", data_pb2.Observation(observed_value=environment_tick_count))]
            )

            async for event in environment_session.event_loop():
                if event.actions:

                    if event.type == cogment.EventType.ACTIVE:
                        action0 = event.actions[0]
                        assert action0.actor_index == 0
                        assert action0.action.action_value == environment_tick_count
                        action1 = event.actions[1]
                        assert action1.actor_index == 1
                        assert action1.action.action_value == environment_tick_count

                    environment_tick_count += 1
                    environment_session.produce_observations(
                        [
                            (
                                "*",
                                data_pb2.Observation(
                                    observed_value=environment_tick_count
                                ),
                            )
                        ]
                    )

        agent_call_count = 0
        agents_tick_count = {}
        agents_ended = {}
        agents_ended["actor_1"] = asyncio.get_running_loop().create_future()
        agents_ended["actor_2"] = asyncio.get_running_loop().create_future()

        async def agent(actor_session):
            nonlocal agent_call_count
            nonlocal agents_tick_count
            nonlocal agents_ended
            agent_call_count += 1
            assert actor_session.name not in agents_tick_count
            assert actor_session.name in agents_ended
            agents_tick_count[actor_session.name] = 0

            actor_session.start()

            async for event in actor_session.event_loop():

                if event.observation:
                    agents_tick_count[actor_session.name] += 1
                    assert (
                        event.observation.snapshot.observed_value ==
                        agents_tick_count[actor_session.name]
                    )
                    actor_session.do_action(
                        data_pb2.Action(
                            action_value=agents_tick_count[actor_session.name]
                        )
                    )

                if event.type == cogment.EventType.ENDING:
                    assert event.observation
                    assert len(event.rewards) == 0
                    assert len(event.messages) == 0

                    assert (
                        event.observation.snapshot.observed_value ==
                        agents_tick_count[actor_session.name]
                    )

            agents_ended[actor_session.name].set_result(True)

        # Explicitely disabling prometheus by not provideing any registry
        context = cogment.Context(
            cog_settings=cog_settings,
            user_id="test_controller_controlled_trial",
            prometheus_registry=None,
        )

        pre_trial_hook = AsyncMock(wraps=create_pre_trial_hook(environment_implementation="test_environment"))
        context.register_pre_trial_hook(pre_trial_hook)
        context.register_environment(impl_name="test_environment", impl=environment)
        context.register_actor(impl_name="test", impl=agent)

        prometheus_port = find_free_port()
        served_endp = cogment.ServedEndpoint(cogment_test_setup["test_port"])
        asyncio.create_task(
            context.serve_all_registered(
                served_endpoint=served_endp, prometheus_port=prometheus_port
            )
        )
        await asyncio.sleep(1)

        with unittest_case.assertRaises(urllib.error.URLError):
            urllib.request.urlopen(f"http://localhost:{prometheus_port}")

        endp = cogment.Endpoint(cogment_test_setup["orchestrator_endpoint"])
        controller = context.get_controller(endpoint=endp)
        logger.info("--starting trial--")
        trial_id = await controller.start_trial(trial_config=data_pb2.TrialConfig())
        logger.info("--trial started--")

        pre_trial_hook.assert_called_once()
        pre_trial_hook.assert_awaited_once()

        await asyncio.sleep(3)

        logger.info("--requesting trial termination--")
        await controller.terminate_trial(trial_id)
        logger.info("--trial termination request sent--")

        await asyncio.wait(agents_ended.values())

        assert environment_call_count == 1
        assert agent_call_count == 2

        assert environment_tick_count > 0

        unittest_case.assertCountEqual(agents_tick_count.keys(), ["actor_1", "actor_2"])
        unittest_case.assertEqual(
            agents_tick_count["actor_1"], agents_tick_count["actor_2"]
        )
        unittest_case.assertGreaterEqual(
            environment_tick_count, agents_tick_count["actor_1"]
        )

        logger.info(f"test_controller_controlled_trial finished")
        await context._grpc_server.stop(grace=5.0)  # To prepare for next test

    @pytest.mark.use_cogment
    @pytest.mark.asyncio
    async def test_actor_order(
        self, cogment_test_setup, unittest_case, cog_settings, data_pb2
    ):
        agents = ["actor_A", "actor_B", "actor_C", "actor_D"]

        async def pre_trial_hook(session):
            host = "grpc://localhost:" + str(cogment_test_setup["test_port"])
            session.actors = [
                {
                    "name": agents[0],
                    "actor_class": "my_actor_class_2",
                    "endpoint": host,
                    "implementation": "basic",
                    "config": None,
                },
                {
                    "name": agents[1],
                    "actor_class": "my_actor_class_1",
                    "endpoint": host,
                    "implementation": "basic",
                    "config": None,
                },
                {
                    "name": agents[2],
                    "actor_class": "my_actor_class_2",
                    "endpoint": host,
                    "implementation": "basic",
                    "config": None,
                },
                {
                    "name": agents[3],
                    "actor_class": "my_actor_class_1",
                    "endpoint": host,
                    "implementation": "basic",
                    "config": None,
                },
            ]

            session.validate()

        async def environment(environment_session):
            actors = environment_session.get_active_actors()
            assert len(actors) == len(agents)
            for i in range(len(agents)):
                assert actors[i].actor_name == agents[i]

            environment_session.start([("*", data_pb2.Observation(observed_value=1))])

            async for event in environment_session.event_loop():
                if event.actions and event.type == cogment.EventType.ACTIVE:
                    for i in range(len(event.actions)):
                        action_i = event.actions[i]
                        assert action_i.actor_index == i  # This may not always be true
                        assert (
                            action_i.action.action_value == action_i.actor_index + 100
                        )

                    if environment_session.get_tick_id() < 3:
                        environment_session.produce_observations(
                            [("*", data_pb2.Observation(observed_value=1))]
                        )
                    elif environment_session.get_tick_id() == 3:
                        environment_session.end(
                            [("*", data_pb2.Observation(observed_value=1))]
                        )

        agents_ended = {}
        for name in agents:
            agents_ended[name] = asyncio.get_running_loop().create_future()

        async def basic_agent(actor_session):
            nonlocal agents_ended
            assert actor_session.name in agents
            assert actor_session.name in agents_ended

            actor_session.start()

            async for event in actor_session.event_loop():
                if event.observation and event.type == cogment.EventType.ACTIVE:
                    agent_index = agents.index(actor_session.name)
                    actor_session.do_action(
                        data_pb2.Action(action_value=agent_index + 100)
                    )

            agents_ended[actor_session.name].set_result(True)

        context = cogment.Context(
            cog_settings=cog_settings,
            user_id="test_actor_order",
            prometheus_registry=None,
        )

        context.register_environment(impl=environment)
        context.register_actor(impl_name="basic", impl=basic_agent)
        context.register_pre_trial_hook(pre_trial_hook)

        served_endp = cogment.ServedEndpoint(cogment_test_setup["test_port"])
        asyncio.create_task(
            context.serve_all_registered(
                served_endpoint=served_endp, prometheus_port=None
            )
        )
        await asyncio.sleep(1)

        endp = cogment.Endpoint(cogment_test_setup["orchestrator_endpoint"])
        controller = context.get_controller(endpoint=endp)
        await controller.start_trial(
            trial_config=data_pb2.TrialConfig(trial_config_value=31)
        )

        await asyncio.wait(agents_ended.values())

        logger.info(f"test_actor_order_trial finished")
        await context._grpc_server.stop(grace=5.0)  # To prepare for next test



# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Tuple, Any
import asyncio
import time
import inspect
import threading

import grpc.aio  # type: ignore

import cogment.api.model_registry_pb2 as model_registry_api

from cogment.errors import CogmentError
from cogment.utils import logger
from cogment.grpc_metadata import GrpcMetadata


GRPC_BYTE_SIZE_LIMIT = 4 * 1024 * 1024  # 4MB

# This should not be a user decision, but we don't want to change the gRPC API at this time.
# Arbitrary size hopefully small enough not to hit the gRPC size limit.
# Unfortunately it is impossible to know without requesting the data, so we play it safe.
_RETRIEVAL_COUNT = 10

# Asyncio based: Not thread safe.
# These could be made thread local, but that would introduce a different set of use-case problems.
_tracked_models = {}  # type: Dict[str, _TrackedModel]
_tracked_models_removal_task = None
_tracked_models_thread_id = None


class ModelIterationInfo:
    def __init__(self, proto_version_info):
        self.model_name = proto_version_info.model_id
        self.iteration = proto_version_info.version_number
        self.timestamp = proto_version_info.creation_timestamp
        self.hash = proto_version_info.data_hash
        self.size = proto_version_info.data_size
        self.stored = proto_version_info.archived
        self.properties = proto_version_info.user_data  # Ideally 'dict(proto_version_info.user_data)' but slower

    def __str__(self):
        result = f"ModelIterationInfo: model_name = {self.model_name}, iteration = {self.iteration}"
        result += f", timestamp = {self.timestamp}, hash = {self.hash}, size = {self.size}"
        result += f", stored = {self.stored}, properties = {self.properties}"
        return result


class ModelInfo:
    def __init__(self, proto_model_info):
        self.name = proto_model_info.model_id
        self.properties = proto_model_info.user_data  # Ideally 'dict(proto_model_info.user_data)' but slower

    def __str__(self):
        result = f"ModelInfo: name = {self.name}, properties = {self.properties}"
        return result


class LatestModel:
    def __init__(self, tracked_model):
        self._model = tracked_model
        self._model.increment_reference()

        # We only provide the name, because the properties can change
        # over time and we don't track them with the iteration info.
        self.name = self._model.model_info.name

    def __del__(self):
        self._model.decrement_reference()

    def is_deserialized(self) -> bool:
        return self._model.deserialized_func is not None

    async def get(self) -> Tuple[Any, ModelIterationInfo]:
        await self.wait_for_available()
        return self.get_no_wait()

    def is_available(self) -> bool:
        return self._model.available.is_set()  # type: ignore

    async def wait_for_available(self) -> None:
        self._model.start_tracking()
        await self._model.available.wait()

    async def wait_for_newer(self, iteration: int) -> None:
        await self.wait_for_available()
        while iteration >= self._model.iteration_info.iteration:
            event = asyncio.Event()
            self._model.set_event_on_new_model(event)
            await event.wait()

    # Undocumented: But we leave it available for special cases or low latency step trials
    def get_no_wait(self) -> Tuple[Any, ModelIterationInfo]:
        return self._model.model, self._model.iteration_info


class _TrackedModel:
    def __init__(self, info, deserialize_func, registry):
        self.model_info = info
        self.registry = registry
        self.deserialize_func = deserialize_func

        self.available = asyncio.Event()
        self.last_ref = time.monotonic()
        self.ref_count = 0

        # The model is a bytes string if 'deserialize_func' is None
        # Otherwise it is the return value of `deserialize_func`
        self.model = None
        self.iteration_info = None

        self._tracking_task = None
        self._new_events = []

    def increment_reference(self):
        self.ref_count += 1
        self.start_tracking()

    def decrement_reference(self):
        self.ref_count -= 1
        if self.ref_count > 0:
            return
        self.last_ref = time.monotonic()
        self.ref_count = 0

    def start_tracking(self):
        if self._tracking_task is None:
            self._tracking_task = asyncio.create_task(self._track_model())
            # self._tracking_task.set_name(f"Cogment-Model [{self.model_info.name}] tracking") # Python 3.8

    def terminate(self):
        if self._tracking_task is not None:
            self._tracking_task.cancel()
            self._tracking_task = None

    def set_event_on_new_model(self, event):
        self._new_events.append(event)

    async def _track_model(self):
        name = self.model_info.name
        registry = self.registry

        # Continuously track/update model
        try:
            async for info in registry.iteration_updates(name):
                registry_model = await registry.retrieve_model(name, info.iteration)
                if registry_model is None:
                    logger.warning(
                        f"Tracked Model [{name}] iteration [{info.iteration}] unavailable:"
                        " it was probably flushed from the model registry cache"
                    )
                    continue
                if self.deserialize_func is not None:
                    self.model = self.deserialize_func(registry_model)
                else:
                    self.model = registry_model
                self.iteration_info = info

                if not self.available.is_set():
                    self.available.set()
                    logger.debug(f"First tracked model [{name}] iteration info [{info}]")

                if len(self._new_events) > 0:
                    running_new_events = self._new_events
                    self._new_events = []
                    for event in running_new_events:
                        event.set()

        except Exception:
            logger.exception(f"Failed to track model [{name}]")

        self._tracking_task = None


async def _tracked_models_removal():
    global _tracked_models_removal_task
    global _tracked_models_thread_id

    _tracked_models_thread_id = threading.get_ident()

    try:
        while True:
            await asyncio.sleep(60)
            now = time.monotonic()

            if len(_tracked_models) == 0:
                self_task = _tracked_models_removal_task
                _tracked_models_removal_task = None
                _tracked_models_thread_id = None
                logger.debug("Cancelled task for tracked models removal")
                self_task.cancel()

            for name in list(_tracked_models):
                if _tracked_models[name].ref_count == 0 and (now - _tracked_models[name].last_ref) > 300:
                    _tracked_models[name].terminate()
                    del _tracked_models[name]
                    logger.debug(f"Stopped tracking model [{name}]. [{len(_tracked_models)}] tracked models left.")

    except Exception:
        logger.exception("Tracked models removal task failure")

    _tracked_models_removal_task = None
    _tracked_models_thread_id = None


class ModelRegistry:
    def __init__(self, stub, endpoint_url, metadata: GrpcMetadata = GrpcMetadata()):
        self._model_registry_stub = stub
        self._endpoint_url = endpoint_url
        self._metadata = metadata.copy()

    def __str__(self):
        return "ModelRegistry"

    def __await__(self):
        """
         Make model registry instances awaitable
        This, frankly dirty hack, allows using `await context.get_model_registry_v2(endpoint)` with any kind of endpoint
        """

        async def _self():
            return self

        return asyncio.create_task(_self()).__await__()

    def has_specs(self):
        return True  # This class does not rely on the spec

    async def store_model(
        self, name: str, model: bytes, iteration_properties: Dict[str, str] = None
    ) -> ModelIterationInfo:
        return await self._send_model(name, model, iteration_properties, True)

    async def publish_model(
        self, name: str, model: bytes, iteration_properties: Dict[str, str] = None
    ) -> ModelIterationInfo:
        return await self._send_model(name, model, iteration_properties, False)

    async def _send_model(self, name, model, iteration_properties, store) -> ModelIterationInfo:
        if model is None and iteration_properties is not None:
            raise CogmentError(f"Cannot send iteration properties with no model iteration")
        if model is not None and len(model) == 0:
            raise CogmentError("Model is empty")

        model_info = await self._get_model_info(name)
        new_model = model_info is None

        if new_model:
            req = model_registry_api.CreateOrUpdateModelRequest()
            req.model_info.model_id = name
            _ = await self._model_registry_stub.CreateOrUpdateModel(req, metadata=self._metadata.to_grpc_metadata())

        if model is None:
            return None

        def generate_chunks():
            try:
                version_info = model_registry_api.ModelVersionInfo()
                version_info.model_id = name
                version_info.archived = store
                version_info.data_size = len(model)
                if iteration_properties is not None:
                    version_info.user_data.update(iteration_properties)
                header = model_registry_api.CreateVersionRequestChunk.Header(version_info=version_info)
                header_chunk = model_registry_api.CreateVersionRequestChunk(header=header)
                yield header_chunk

                max_size = GRPC_BYTE_SIZE_LIMIT // 2
                for index in range(0, len(model), max_size):
                    body = model_registry_api.CreateVersionRequestChunk.Body()
                    body.data_chunk = model[index : index + max_size]
                    body_chunk = model_registry_api.CreateVersionRequestChunk(body=body)
                    yield body_chunk

            except Exception as exc:
                raise CogmentError(f"Failure to generate model data for sending [{exc}]")

        reply = await self._model_registry_stub.CreateVersion(
            generate_chunks(), metadata=self._metadata.to_grpc_metadata()
        )

        return ModelIterationInfo(reply.version_info)

    async def retrieve_model(self, name: str, iteration: int = -1) -> bytes:
        req = model_registry_api.RetrieveVersionDataRequest()
        req.model_id = name
        req.version_number = iteration

        model = bytes()
        try:
            async for chunk in self._model_registry_stub.RetrieveVersionData(
                req, metadata=self._metadata.to_grpc_metadata()
            ):
                model += chunk.data_chunk
        except Exception as exc:
            # Either the model does not exist, the iteration does not exist, or there was a real error.
            # The Model Registry should be fixed to differentiate
            logger.debug(f"Failed to retrieve iteration [{iteration}] for model [{name}]: [{exc}]")
            model = None

        if model is None:
            model_info = await self._get_model_info(name)
            if model_info is None:
                raise CogmentError(f"Unknown model [{name}]")
            else:
                return None  # Model exists, but not the iteration (probably)

        return model

    async def remove_model(self, name: str) -> None:
        req = model_registry_api.DeleteModelRequest(model_id=name)
        _ = await self._model_registry_stub.DeleteModel(req, metadata=self._metadata.to_grpc_metadata())

    async def list_models(self) -> List[ModelInfo]:
        req = model_registry_api.RetrieveModelsRequest(models_count=_RETRIEVAL_COUNT)
        models = []
        try:
            while True:
                reply = await self._model_registry_stub.RetrieveModels(req, metadata=self._metadata.to_grpc_metadata())
                for model_info in reply.model_infos:
                    models.append(ModelInfo(model_info))

                # There is a bug in model registry and we can't rely on "next_model_handle" to know the end;
                # this forces us to make one extra request if the number is a multiple of the count!
                if len(reply.model_infos) < _RETRIEVAL_COUNT:
                    break
                req.model_handle = reply.next_model_handle

        except Exception as exc:
            raise CogmentError(f"Failed to retrieve model list [{exc}]")

        return models

    async def list_iterations(self, model_name: str) -> List[ModelIterationInfo]:
        req = model_registry_api.RetrieveVersionInfosRequest(model_id=model_name, versions_count=_RETRIEVAL_COUNT)
        iterations = []
        try:
            while True:
                reply = await self._model_registry_stub.RetrieveVersionInfos(
                    req, metadata=self._metadata.to_grpc_metadata()
                )
                for iteration_info in reply.version_infos:
                    iterations.append(ModelIterationInfo(iteration_info))

                # There is a bug in model registry and we can't rely on "next_version_handle" to know the end;
                # this forces us to make one extra request if the number is a multiple of the count!
                if len(reply.version_infos) < _RETRIEVAL_COUNT:
                    break
                req.version_handle = reply.next_version_handle

        except Exception as exc:
            raise CogmentError(f"Failed to retrieve model [{model_name}] iteration list: [{exc}]")

        return iterations

    async def get_iteration_info(self, name: str, iteration: int) -> ModelIterationInfo:
        req = model_registry_api.RetrieveVersionInfosRequest(model_id=name, version_numbers=[iteration])
        try:
            reply = await self._model_registry_stub.RetrieveVersionInfos(
                req, metadata=self._metadata.to_grpc_metadata()
            )
        except Exception as exc:
            # Either the model does not exist, the iteration does not exist, or there was a real error.
            # The Model Registry should be fixed to differentiate.
            logger.debug(f"Failed to retrieve iteration [{iteration}] info for model [{name}]: [{exc}]")
            reply = None

        if reply is None:
            model_info = await self._get_model_info(name)
            if model_info is None:
                raise CogmentError(f"Unknown model [{name}]")
            else:
                return None  # Model exists, but not the iteration (probably)

        if len(reply.version_infos) == 0:
            return None
        if len(reply.version_infos) > 1:
            logger.warning(f"Model Registry unexpectedly returned multiple iterations [{len(reply.version_infos)}]")

        return ModelIterationInfo(reply.version_infos[0])

    async def update_model_info(self, name: str, properties: Dict[str, str]) -> None:
        req = model_registry_api.CreateOrUpdateModelRequest()
        req.model_info.model_id = name
        req.model_info.user_data.update(properties)
        try:
            _ = await self._model_registry_stub.CreateOrUpdateModel(req, metadata=self._metadata.to_grpc_metadata())
        except Exception as exc:
            raise CogmentError(f"Failed to store model [{name}] properties: [{exc}]")

    async def _get_model_info(self, name):
        req = model_registry_api.RetrieveModelsRequest(model_ids=[name])
        try:
            reply = await self._model_registry_stub.RetrieveModels(req, metadata=self._metadata.to_grpc_metadata())
        except Exception as exc:
            # We assume any error is due to the model not existing!
            logger.debug(f"Failed to retrieve model [{name}] info: [{exc}]")
            return None

        if len(reply.model_infos) == 0:
            logger.debug(f"No model [{name}]")
            return None
        if len(reply.model_infos) > 1:
            logger.warning(f"Model Registry unexpectedly returned multiple model infos [{len(reply.model_infos)}]")

        return reply.model_infos[0]

    async def get_model_info(self, name: str) -> ModelInfo:
        model_info = await self._get_model_info(name)
        if model_info is None:
            return None
        result = ModelInfo(model_info)

        return result

    async def iteration_updates(self, model_name: str):
        req = model_registry_api.VersionUpdateRequest()
        req.model_id = model_name

        reply_itor = self._model_registry_stub.VersionUpdate(req, metadata=self._metadata.to_grpc_metadata())
        if not reply_itor:
            raise CogmentError("Failed to connect to the Model Registry")

        try:
            async for update in reply_itor:
                if update == grpc.aio.EOF:
                    raise CogmentError(f"No response to Model Registry iteration update request for [{model_name}]")
                yield ModelIterationInfo(update.version_info)

        except grpc.aio.AioRpcError as exc:
            logger.debug(f"gRPC failed status details: [{exc.debug_error_string()}]")
            if exc.code() == grpc.StatusCode.UNAVAILABLE:
                logger.error(f"Model Registry communication lost [{exc.details()}]")
            else:
                logger.exception("Model Registry communication -- Unexpected aio failure")
                raise

    # Utility function for simple use cases. More complex cases must use 'iteration_update' explicitly.
    # This may fail if called in different threads or with different async loops.
    async def track_latest_model(
        self,
        name: str,
        deserialize_func: Callable[[bytes], Any] = None,
        initial_wait: int = 0,
    ) -> LatestModel:
        global _tracked_models_removal_task

        if name not in _tracked_models:
            _tracked_models[name] = None

            # We want to fail early if 'deserialize_func' is wrong
            if deserialize_func is not None:
                sig = inspect.signature(deserialize_func)
                if len(sig.parameters) != 1:
                    raise CogmentError(f"Deserialize function must take only one parameter [{len(sig.parameters)}]")

            # Check and wait for model to exist
            model_info = await self.get_model_info(name)
            fail_count = 0
            while model_info is None:
                fail_count += 1
                if fail_count > initial_wait:
                    del _tracked_models[name]
                    raise CogmentError(f"Model [{name}] not found in Model Registry")
                logger.debug(f"Model [{name}] not yet in Model Registry")
                await asyncio.sleep(1.0)
                model_info = await self.get_model_info(name)

            _tracked_models[name] = _TrackedModel(model_info, deserialize_func, self)

        else:  # Other task is already requesting the initial model info: wait
            while _tracked_models[name] is None:
                await asyncio.sleep(1.0)
                if name not in _tracked_models:
                    raise CogmentError(f"Model [{name}] may not be found in Model Registry")

        tracked_model = _tracked_models[name]
        if deserialize_func is not None and tracked_model.deserialize_func != deserialize_func:
            raise CogmentError(f"Deserialize function mismatch with already set function")
        if tracked_model.registry._endpoint_url != self._endpoint_url:
            raise CogmentError(
                f"Different model registry to track the same model [{name}]"
                f": [{tracked_model.registry._endpoint_url}] vs [{self._endpoint_url}]"
            )

        if _tracked_models_removal_task is None:
            _tracked_models_removal_task = asyncio.create_task(_tracked_models_removal())
            # _tracked_models_removal_task.set_name("Cogment-Stop tracking unused models") # Python 3.8
        elif _tracked_models_thread_id != threading.get_ident():
            logger.warning("Tracking global models on different threads")

        return LatestModel(tracked_model)



# Copyright 2023 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc.aio  # type: ignore

import cogment.api.datalog_pb2_grpc as datalog_grpc_api
import cogment.api.common_pb2 as common_api
import cogment.api.datalog_pb2 as datalog_api

from cogment.control import TrialState
from cogment.datalog import DatalogSession, LogParams
from cogment.parameters import TrialParameters
from cogment.errors import CogmentError
import cogment.utils as utils
from cogment.utils import logger
from cogment.session import RecvObservation, ActorStatus, RecvAction, RecvMessage, RecvReward

import asyncio


class LogSample:
    """Class representing a trial sample for the datalog service."""

    def __init__(self, params):
        self.tick_id = None
        self.out_of_sync = None
        self.timestamp = None
        self.state = None
        self.events = None

        self._raw_sample = None
        self._actor_indexes = None

        self._log_params = None
        self._parameters = None
        if type(params) == LogParams:
            logger.deprecated(f"Deprecated use of LogParams")
            self._log_params = params
            self._nb_actors = params.nb_actors
        elif type(params) == TrialParameters:
            self._parameters = params
            self._nb_actors = len(params.actors)
        else:
            raise CogmentError(f"Wrong type of params provided [{type(params)}]")

    def __str__(self):
        result = f"LogSample: {self._raw_sample}"
        return result

    # Type of serialized data being produced and consumed by this class.
    # This is dependent on all the underlying protobuf messages used to
    # serialize/deserialize, and should be incremented if any of them changes in
    # a backward or forward incompatible way. API 1.0 could be considered a type 0 or 1.
    # Current dependencies: DatalogSample, SampleInfo, TrialState, ObservationSet,
    #                       Action, Reward, RewardSource, Message
    def get_serialization_type(self):
        return 2

    def _set(self, raw_sample):
        self._set_from_sample(raw_sample)

    def serialize(self):
        if self._raw_sample is None:
            raise CogmentError("Not set, nothing to serialize")
        return self._raw_sample.SerializeToString()

    def deserialize(self, raw_string):
        sample = datalog_api.DatalogSample()
        sample.ParseFromString(raw_string)
        self._set_from_sample(sample)

    def _set_from_sample(self, sample):
        if type(sample) != datalog_api.DatalogSample:
            raise CogmentError(f"Wrong type of sample provided [{type(sample)}]")
        if sample.HasField("info"):
            self.state = TrialState(sample.info.state)
            self.out_of_sync = sample.info.out_of_sync
            self.tick_id = sample.info.tick_id
            self.timestamp = sample.info.timestamp
            self.events = sample.info.special_events
        else:
            self.state = None
            self.out_of_sync = None
            self.tick_id = None
            self.timestamp = None
            self.events = None

        self._raw_sample = sample

    def _get_actor_index(self, actor_name):
        if self._actor_indexes is None:
            indexes = {}
            if self._log_params:
                for index, actor in enumerate(self._log_params._raw_params.actors):
                    indexes[actor.name] = index
            else:
                for index, actor in enumerate(self._parameters.actors):
                    indexes[actor.name] = index
            self._actor_indexes = indexes

        return self._actor_indexes.get(actor_name)

    def all_actor_names(self):
        if self._log_params:
            for index in range(self._nb_actors):
                yield self._log_params.get_actor_name(index)
        else:
            for actor in self._parameters.actors:
                yield actor.name

    def get_action(self, actor):
        if isinstance(actor, int):
            actor_index = actor
        elif isinstance(actor, str):
            actor_index = self._get_actor_index(actor)
        else:
            raise CogmentError(f"Wrong type of actor parameter [{type(actor)}]: must be int or str")
        if actor_index is None or actor_index < 0 or actor_index >= self._nb_actors:
            raise CogmentError(f"Invalid actor [{actor}] [{actor_index}]")

        if len(self._raw_sample.actions) > 0:
            data = self._raw_sample.actions[actor_index]
            if actor_index in self._raw_sample.unavailable_actors:
                status = ActorStatus.UNAVAILABLE
                timestamp = 0
                action_space = None
            elif actor_index in self._raw_sample.default_actors:
                status = ActorStatus.DEFAULT
                timestamp = 0
                action_space = None
            else:
                status = ActorStatus.ACTIVE
                timestamp = data.timestamp
                if self._log_params:
                    action_space = self._log_params.get_actor(actor_index)["action_space"]()
                else:
                    action_space = self._parameters.actors[actor_index].actor_class_spec.action_space()
                action_space.ParseFromString(data.content)

            return RecvAction(actor_index, data.tick_id, status, timestamp, action_space)

        else:
            return None

    def get_observation(self, actor):
        if isinstance(actor, int):
            actor_index = actor
        elif isinstance(actor, str):
            actor_index = self._get_actor_index(actor)
        else:
            raise CogmentError(f"Wrong type of actor parameter [{type(actor)}]: must be int or str")
        if actor_index is None or actor_index < 0 or actor_index >= self._nb_actors:
            raise CogmentError(f"Invalid actor [{actor}] [{actor_index}]")

        if self._log_params:
            obs_space = self._log_params.get_actor(actor_index)["observation_space"]()
        else:
            obs_space = self._parameters.actors[actor_index].actor_class_spec.observation_space()

        if self._raw_sample.HasField("observations"):
            data = self._raw_sample.observations
            obs_index = data.actors_map[actor_index]
            obs_content = data.observations[obs_index]
            obs_space.ParseFromString(obs_content)
            return RecvObservation(data, obs_space)
        else:
            return None

    def all_rewards(self):
        for rew in self._raw_sample.rewards:
            yield RecvReward(rew)

    def all_messages(self):
        for msg in self._raw_sample.messages:
            yield RecvMessage(msg)


async def _read_sample(context, session, settings):
    try:
        while True:
            request = await context.read()

            if request == grpc.aio.EOF:
                logger.info(f"The orchestrator disconnected from LogExpoterService.")
                break

            elif request.HasField("sample"):
                trial_ended = (request.sample.info.state == common_api.TrialState.ENDED)
                sample = LogSample(session.trial_parameters)
                sample._set(request.sample)
                session._new_sample(sample)
                if trial_ended:
                    logger.debug("Last log sample received for trial")
                    break
            else:
                logger.warning(f"Invalid request received from the orchestrator : {request}")

    except asyncio.CancelledError as exc:
        logger.debug(f"DatalogServicer '_read_sample' coroutine cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_read_sample")
        raise

    # Exit the loop
    session._new_sample(None)


class DatalogServicer(datalog_grpc_api.DatalogSPServicer):
    """Internal datalog servicer class."""

    def __init__(self, impl, cog_settings):
        self._impl = impl
        self._cog_settings = cog_settings
        logger.info("Datalog Service started")

    # Override
    async def RunTrialDatalog(self, request_iterator, context):
        reader_task = None
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]
            user_id = metadata["user-id"]

            request = await context.read()

            if not request.HasField("trial_params"):
                raise CogmentError(f"Initial data log request for [{trial_id}] does not contain parameters.")

            trial_parameters = TrialParameters(self._cog_settings, raw_params=request.trial_params)

            session = DatalogSession(self._impl, trial_id, user_id, trial_parameters)
            user_task = session._start_user_task()

            reader_task = asyncio.create_task(_read_sample(context, session, self._cog_settings))

            normal_return = await user_task

            if normal_return:
                logger.debug(f"User datalog implementation returned")
            else:
                logger.debug(f"User datalog implementation was cancelled")

        except asyncio.CancelledError as exc:
            logger.debug(f"Datalog implementation coroutine cancelled: [{exc}]")

        except Exception:
            logger.exception("RunTrialDatalog")
            raise

        finally:
            if reader_task is not None:
                reader_task.cancel()

    # Override
    async def Version(self, request, context):
        try:
            return utils.list_versions()
        except Exception:
            logger.exception("Version")
            raise

    # Override
    async def Status(self, request, context):
        reply = common_api.StatusReply()

        try:
            if len(request.names) == 0:
                return reply

            # We purposefully don't scan for "*" ahead of time to allow explicit values before
            all = False
            for name in request.names:
                if name == "*":
                    all = True

                if all or name == "overall_load":
                    load = utils.machine_load()
                    if load:
                        reply.statuses["overall_load"] = str(load)
                    else:
                        reply.statuses["overall_load"] = "0"

                if all:
                    break

        except Exception:
            logger.exception("Status")
            raise

        return reply
